import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.io.*;
import no.uib.cipr.matrix.distributed.*;
import no.uib.cipr.matrix.*;

import controlP5.*;
ControlP5 cP5;
ControlWindow cWindow;

DenseVector a, aNext;
DenseVector b, bNext;

FlexCompRowMatrix LA;
FlexCompRowMatrix LB;
IterativeSolver solverA;
IterativeSolver solverB;

int w = 300;
int h = 300;

float delx = 1.0; //.6
float delxSq = delx*delx;
float deltaT = .009;

//decreasing reactionRate is the same as increasing the diffusion rate
float reactionRate = .5;//2.0

//settings based on http://hopf.chem.brandeis.edu/yanglingfa/pattern/bruss/index.html
//diffusion rates
float aRate = 5;
float aRateX = aRate;//2.86422
float aRateY = aRate;
float bRate = 12;//10.0
//alphaR is the rate of production of A
//betaR is the rate of decay of A into B
float alphaR = 2.5;//1.5
float betaR = 8;//3.25


//not used in brusselator
float F = .04;
float k = -.05;
float Farr[] = new float[w*h];
float karr[] = new float[w*h];

float EPSILON = .005;

int iNext[] = new int[w];
int jNext[] = new int[h];
int iPrev[] = new int[w];
int jPrev[] = new int[h];

boolean movie = false;
String name = ""+day()+hour()+minute()+second();

double minA = 999;
double maxA = -999;
void setup() {
  size(w,h);
  colorMode(HSB,1.0);
  a = new DenseVector(w*h);
  b = new DenseVector(w*h);
  
  LA = new FlexCompRowMatrix(w*h,w*h);
  LB = new FlexCompRowMatrix(w*h,w*h);
  solverA = new CGS(a);
  solverB = new CGS(b);

  for(int i=0;i<w;++i) {
    iNext[i] = (i+1)%w;
    iPrev[i] = (i-1+w)%w;
  }
  for(int j=0;j<h;++j) {
    jNext[j] = (j+1)%h;
    jPrev[j] = (j-1+h)%h;
  }
  setupFK();
  setupAB();
  println("initialize");
  initialize();
   println("compute laplacian");
  computeLaplacian();
  println("load operators");
  
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  
   
  gui();
  
}

void initialize() {
  for(int i=0;i<w*h;++i) {    
    a.set(i,alphaR+random(-5,5));
    b.set(i,betaR/alphaR+random(-5,5));
    //a.set(i,.5+random(-.01,.01));//random(-1,1.0));
    //b.set(i,.25+random(-.01,.01));//random(-1,1.0));
  }
   for(int num=0;num<1;++num) {
    int x = int(random(w));
    int y = int(random(h));
//    setBall(x,y,8,1.0,0.0);
  //  setBall(x,y,10,1,3);
  }
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
}

void draw() {
  cP5.draw();
  background(1.0);
  noStroke();
  long start = millis();
  while(millis()-start < 30) step();
  loadPixels();
    float valA;
   // beginShape(QUADS);
    int index = 0;
    for(int i=0;i<w;++i) {
      for(int j=0;j<h;++j) {
          valA = (float)a.get(index);
          float colorific = (float)((valA-minA)/(maxA-minA));
          pixels[index++] = color(colorific,1.0,1.0);
      }
    }
  updatePixels();
    //endShape();
}

void step() {
  //long start = millis();
  diffusion();
  //println(millis()-start);
  reaction();
}

void diffusion() {
  float valA, valB;
  LA.multAdd(1,a,a);
  LB.multAdd(1,b,b);
  
  //for(int i=0;i<2;++i) {
    //gaussSeidel((FlexCompRowMatrix)LA, aNext, a);
    //gaussSeidel((FlexCompRowMatrix)LB, bNext, b);
  //}
  try {
    //solverA.solve(LA,a,a);
    //solverB.solve(LB,b,b);
  } catch(Exception e) {
    println("NO CONVERGE");
  }
  
  
  /*
  DenseVector temp = a;
  a = aNext;
  aNext = temp;
  temp = b;
  b = bNext;
  bNext = temp;
  */
}

void reaction() {
  minA = 999;
  maxA = -999;
  double valA, valB;
  double currA, currB;
  int index = 0;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      currA = a.get(index);
      currB = b.get(index);
      valA = deltaT*reactionA(currA,currB,index);
      valB = deltaT*reactionB(currA,currB,index);
      a.add(index,valA);
      b.add(index,valB);
      index++;
      if(currA+valA > maxA) maxA = currA+valA;
      if(currA+valA < minA) minA = currA+valA;
    }
  }
}

double reactionA(double aVal, double bVal,int i) {
  return reactionABrussel(aVal,bVal,i);
}

double reactionB(double aVal, double bVal,int i) {
  return reactionBBrussel(aVal,bVal,i);
}

void setBall(int x, int y, float dist, float aVal, float bVal) {
  int m = int(dist);
  int cx,cy;
  int index;
  for(int i=-m;i<=m;++i) {
    for(int j=-m;j<=m;++j) {
      if(i*i+j*j < dist*dist) {
        cx = (x+i+w)%w; 
        cy = (y+j+h)%h; 
        index = xyToIndex(cx,cy);
        a.set(index,aVal+random(-.01,.01));
        b.set(index,bVal+random(-.01,.01));
      }      
    }
  }
}

void keyPressed() {
  if(key == ' ') {
    step();
  } else if(key == 's') {
    save("reaction_"+name+".png");
  }
}

void mousePressed() {
  println(a.get(mouseX+mouseY*w) + " " + b.get(mouseX+mouseY*w));
}

