import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.io.*;
import no.uib.cipr.matrix.distributed.*;
import no.uib.cipr.matrix.*;

DenseVector a, aNext;
DenseVector b, bNext;

Matrix LA;
Matrix LB;
IterativeSolver solverA;
IterativeSolver solverB;

//float[][][] a;
//float[][][] b;
//float[][][] aNext;
//float[][][] bNext;
int w = 256;
int h = 256;

float delx = .6; //.6
float delxSq = delx*delx;
float deltaT = 1.0;
float reactionRate = .0009;//0.003 //1.0/128.0; //Turing 1/128.0
float aRateX = 0.00028;//.04 //.0028 //.1 //.2 //1/16:1/14 OR 1/16:1
float aRateY = 0.00028;//.04
float bRate = 0.05;//.5  //.05//.1 //.25
float F = .04;//.635 //.064//.04 //.06
float k = -.05;//.61 // .0606.06 //.062
float Farr[] = new float[w*h];
float karr[] = new float[w*h];

float EPSILON = .005;

int iNext[] = new int[w];
int jNext[] = new int[h];
int iPrev[] = new int[w];
int jPrev[] = new int[h];

boolean movie = false;
String name = ""+day()+hour()+minute()+second();

double minA = 999;
double maxA = -999;
void setup() {
  size(w,h);
  colorMode(HSB,1.0);
  a = new DenseVector(w*h);
  b = new DenseVector(w*h);
  
  LA = new FlexCompRowMatrix(w*h,w*h);
  LB = new FlexCompRowMatrix(w*h,w*h);
  solverA = new CGS(a);
  solverB = new CGS(b);

  for(int i=0;i<w;++i) {
    iNext[i] = (i+1)%w;
    iPrev[i] = (i-1+w)%w;
  }
  for(int j=0;j<h;++j) {
    jNext[j] = (j+1)%h;
    jPrev[j] = (j-1+h)%h;
  }
  setupFK();
  setupAB();
  println("initialize");
  initialize();
  println("load operators");
  
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  
}

void initialize() {
  for(int i=0;i<w*h;++i) {
    //a.set(i,.5+random(-.01,.01));
    //b.set(i,.25+random(-.01,.01));
    a.set(i,0);//random(-1,1.0));
    b.set(i,0);//random(-1,1.0));
  }
  for(int num=0;num<1;++num) {
    int x = int(random(w));
    int y = int(random(h));
//    setBall(x,y,8,1.0,0.0);
    setBall(x,y,10,0,0);
  }
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  println("compute laplacian");
  computeLaplacian();
}

void draw() {
  background(1.0);
  noStroke();
  long start = millis();
  while(millis()-start < 30) step();
  loadPixels();
    float valA;
   // beginShape(QUADS);
    int index = 0;
    for(int i=0;i<w;++i) {
      for(int j=0;j<h;++j) {
          valA = (float)a.get(index);
          pixels[index++] = color(1.0,0.0,(float)((valA-minA)/(maxA-minA)));
      }
    }
  updatePixels();
    //endShape();
}

void step() {
  //long start = millis();
  diffusion();
  //println(millis()-start);
  reaction();
}

void diffusion() {
  float valA, valB;
  LA.multAdd(1,a,a);
  LB.multAdd(1,b,b);
  
  //for(int i=0;i<2;++i) {
    //gaussSeidel((FlexCompRowMatrix)LA, aNext, a);
    //gaussSeidel((FlexCompRowMatrix)LB, bNext, b);
  //}
  try {
    //solverA.solve(LA,a,a);
    //solverB.solve(LB,b,b);
  } catch(Exception e) {
    println("NO CONVERGE");
  }
  
  
  /*
  DenseVector temp = a;
  a = aNext;
  aNext = temp;
  temp = b;
  b = bNext;
  bNext = temp;
  */
}

void reaction() {
  minA = 999;
  maxA = -999;
  double valA, valB;
  double currA;
  int index = 0;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      currA = a.get(index);
      valA = deltaT*reactionA(currA,b.get(index),index);
      valB = deltaT*reactionB(currA,b.get(index),index);
      a.add(index,valA);
      b.add(index,valB);
      index++;
      if(currA+valA > maxA) maxA = currA+valA;
      if(currA+valA < minA) minA = currA+valA;
    }
  }
}


double reactionA(double aVal, double bVal,int i) {
  return reactionAFHN(aVal,bVal,i);
}

double reactionB(double aVal, double bVal,int i) {
  return reactionBFHN(aVal,bVal,i);
}

void setBall(int x, int y, float dist, float aVal, float bVal) {
  int m = int(dist);
  int cx,cy;
  int index;
  for(int i=-m;i<=m;++i) {
    for(int j=-m;j<=m;++j) {
      if(i*i+j*j < dist*dist) {
        cx = (x+i+w)%w; 
        cy = (y+j+h)%h; 
        index = cx*h+cy;
        a.set(index,aVal+random(-1,1));
        b.set(index,bVal+random(-1,1));
      }      
    }
  }
}

void keyPressed() {
  if(key == ' ') {
    step();
  } else if(key == 's') {
    save("reaction_"+name+".png");
  }
}

