import processing.pdf.*;

import blobDetection.*;
import processing.opengl.*;

import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.io.*;
import no.uib.cipr.matrix.distributed.*;
import no.uib.cipr.matrix.*;

DenseVector a, aNext;
DenseVector b, bNext;

Matrix LA;
Matrix LB;
IterativeSolver solverA;
IterativeSolver solverB;

//float[][][] a;
//float[][][] b;
//float[][][] aNext;
//float[][][] bNext;
int w = 256;
int h = 256;

float delx = .4; //.6
float delxSq = delx*delx;
float deltaT = .25;
float reactionRate = .5;//0.003 //1.0/128.0; //Turing 1/128.0
float aRateX = .04;//.04 //.0028 //.1 //.2 //1/16:1/14 OR 1/16:1
float aRateY = .04;//.04
float bRate = .01;//.5  //.05//.1 //.25
float F = .06;//.635 //.064//.04 //.06
float k = .06;//.61 // .0606.06 //.062
float Farr[] = new float[w*h];
float karr[] = new float[w*h];

float EPSILON = .005;

int iNext[] = new int[w];
int jNext[] = new int[h];
int iPrev[] = new int[w];
int jPrev[] = new int[h];

boolean movie = false;
String name = ""+day()+hour()+minute()+second();

double minA = 999;
double maxA = -999;

BlobDetection bd = new BlobDetection(w,h);
boolean view = false;
void setup() {
  colorMode(RGB,1.0f);
  size(w,h);
  a = new DenseVector(w*h);
  b = new DenseVector(w*h);

  bd.setPosDiscrimination(false);
  bd.setThreshold(.38f);

  LA = new FlexCompRowMatrix(w*h,w*h);
  LB = new FlexCompRowMatrix(w*h,w*h);
  solverA = new CGS(a);
  solverB = new CGS(b);

  for(int i=0;i<w;++i) {
    iNext[i] = (i+1)%w;
    iPrev[i] = (i-1+w)%w;
  }
  for(int j=0;j<h;++j) {
    jNext[j] = (j+1)%h;
    jPrev[j] = (j-1+h)%h;
  }
  println("initialize");
  initialize();
  println("load operators");
  setupFK();

  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
}

void initialize() {
  for(int i=0;i<w*h;++i) {
    //a.set(i,.5+random(-.01,.01));
    //b.set(i,.25+random(-.01,.01));
    a.set(i,.6+random(-.01,.01));
    b.set(i,.25+random(-.01,.01));
  }
  for(int l=0;l<20;++l) {
    int x = int(random(w));
    int y = int(random(h));
    //    setBall(x,y,8,1.0,0.0);
    setBall(x,y,10,-.6,-.25);
  }
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  println("compute laplacian");
  computeLaplacian();
}

void draw() {
  background(1.0);
  noStroke();
  long start = millis();
  while(millis()-start < 30) step();
  loadPixels();
  float valA;
  // beginShape(QUADS);
  int index = 0;
  for(int j=0;j<h;++j) {
    for(int i=0;i<w;++i) {
      valA = (float)a.get(index);
      pixels[index++] = color((float)((valA-minA)/(maxA-minA)));
    }
  }
  updatePixels();
  if(view) {
    background(1.0);
    bd.computeBlobs(pixels);
    drawBlobs();
  }
  //endShape();
}

void drawBlobs() {
  noFill();
  Blob b;
  EdgeVertex eA,eB;
  strokeWeight(1);
  stroke(0);
  for (int n=0 ; n<bd.getBlobNb() ; n++) {
    b=bd.getBlob(n);
    if (b!=null) {
      // Edges
      for (int m=0;m<b.getEdgeNb();m++) {
        eA = b.getEdgeVertexA(m);
        eB = b.getEdgeVertexB(m);
        if (eA !=null && eB !=null)
          line(eA.x*w, eA.y*h, 
          eB.x*w, eB.y*h);
      }
    }
  }
}

void step() {
  //long start = millis();
  diffusion();
  //println(millis()-start);
  reaction();
}

void diffusion() {
  float valA, valB;
  LA.multAdd(1,a,a);
  LB.multAdd(1,b,b);

  //for(int i=0;i<2;++i) {
  //gaussSeidel((FlexCompRowMatrix)LA, aNext, a);
  //gaussSeidel((FlexCompRowMatrix)LB, bNext, b);
  //}
  try {
    //solverA.solve(LA,a,a);
    //solverB.solve(LB,b,b);
  } 
  catch(Exception e) {
    println("NO CONVERGE");
  }


  /*
  DenseVector temp = a;
   a = aNext;
   aNext = temp;
   temp = b;
   b = bNext;
   bNext = temp;
   */
}

void reaction() {
  minA = 999;
  maxA = -999;
  double valA, valB;
  double currA;
  int index = 0;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      currA = a.get(index);
      valA = deltaT*reactionA(currA,b.get(index),index);
      valB = deltaT*reactionB(currA,b.get(index),index);
      a.add(index,valA);
      b.add(index,valB);
      index++;
      if(currA+valA > maxA) maxA = currA+valA;
      if(currA+valA < minA) minA = currA+valA;
    }
  }
}


double reactionA(double aVal, double bVal,int i) {
  return reactionAGS(aVal,bVal,i);
}

double reactionB(double aVal, double bVal,int i) {
  return reactionBGS(aVal,bVal,i);
}

void setBall(int x, int y, float dist, float aVal, float bVal) {
  int m = int(dist);
  int cx,cy;
  int index;
  for(int i=-m;i<=m;++i) {
    for(int j=-m;j<=m;++j) {
      if(i*i+j*j < dist*dist) {
        cx = (x+i+w)%w; 
        cy = (y+j+h)%h; 
        index = xyToIndex(cx,cy);
        a.set(index,aVal+random(-.01,.01));
        b.set(index,bVal+random(-.01,.01));
      }
    }
  }
}

void keyPressed() {
  if(key == ' ') {
    step();
  } 
  else if(key == 's') {
    PGraphics pg = createGraphics(width,height,PDF, "reaction_"+name+".pdf");
    beginRecord(pg);
    loadPixels();
    drawBlobs();
    endRecord();
    save("reaction_"+name+".png");
  } 
  else if(key == 'v') {
    view = !view;
  }
}

