import processing.opengl.*;

import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.volume.*;

import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.io.*;
import no.uib.cipr.matrix.distributed.*;
import no.uib.cipr.matrix.*;

DenseVector a, aNext;
DenseVector b, bNext;

Matrix LA;
Matrix LB;
IterativeSolver solverA;
IterativeSolver solverB;

//float[][][] a;
//float[][][] b;
//float[][][] aNext;
//float[][][] bNext;
int w = 140;
int h = 256;
int d = 140;

float delx = .4; //.6
float delxSq = delx*delx;
float deltaT = .25;
float reactionRate = .6;//0.003 //1.0/128.0; //Turing 1/128.0
float aRateX = .03;//.04 //.0028 //.1 //.2 //1/16:1/14 OR 1/16:1
float aRateY = .03;//.04
float bRate = .01;//.5  //.05//.1 //.25
float F = .04;//.635 //.064//.04 //.06
float k = .06;//.61 // .0606.06 //.062
float Farr[] = new float[w*h];
float karr[] = new float[w*h];

float EPSILON = .005;

int iNext[] = new int[w];
int jNext[] = new int[h];
int iPrev[] = new int[w];
int jPrev[] = new int[h];

boolean movie = false;
String name = ""+day()+hour()+minute()+second();

double minA = 999;
double maxA = -999;
int time = 0;

//toxi marching cubes stuff
VolumetricSpace volume;
IsoSurface surface;
TriangleMesh mesh = new TriangleMesh("why");
float threshold = .5;

boolean view = false;
void setup() {
  colorMode(RGB,1.0f);
  size(w,h);
  a = new DenseVector(w*h);
  b = new DenseVector(w*h);

  LA = new FlexCompRowMatrix(w*h,w*h);
  LB = new FlexCompRowMatrix(w*h,w*h);
  solverA = new CGS(a);
  solverB = new CGS(b);

  for(int i=0;i<w;++i) {
    iNext[i] = (i+1)%w;
    iPrev[i] = (i-1+w)%w;
  }
  for(int j=0;j<h;++j) {
    jNext[j] = (j+1)%h;
    jPrev[j] = (j-1+h)%h;
  }
  println("initialize");
  initialize();
  println("load operators");
  setupFK();
  
  setupVolume();
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
}

void initialize() {
  for(int i=0;i<w*h;++i) {
    //a.set(i,.5+random(-.01,.01));
    //b.set(i,.25+random(-.01,.01));
    a.set(i,.5+random(-.01,.01));
    b.set(i,.25+random(-.01,.01));
  }
  for(int l=0;l<0;++l) {
    int x = int(random(w));
    int y = int(random(h));
    //    setBall(x,y,8,1.0,0.0);
    setBall(x,y,8,1.0,0);
  }
  setBall(w/2,h/2,30,1.0,0);
  //setBall(w/2,h/2+25,10,.5,.25);
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  println("compute laplacian");
  computeLaplacian();
}

void draw() {
  background(1.0);
  noStroke();
  long start = millis();
  //while(millis()-start < 30) step();
  for(int i=0;i<10;++i) step();
  loadPixels();
  float valA;
  // beginShape(QUADS);
  int index = 0;
  for(int j=0;j<h;++j) {
    for(int i=0;i<w;++i) {
      valA = (float)a.get(index);
      pixels[index++] = color((float)((valA-minA)/(maxA-minA)));
    }
  }
  updatePixels();
  if(frameCount%5==0 && time < d) {
    addFrame();
    time++;
  }
  //endShape();
}


void step() {
  //long start = millis();
  diffusion();
  //println(millis()-start);
  reaction();
}

void diffusion() {
  float valA, valB;
  LA.multAdd(1,a,a);
  LB.multAdd(1,b,b);

  //for(int i=0;i<2;++i) {
  //gaussSeidel((FlexCompRowMatrix)LA, aNext, a);
  //gaussSeidel((FlexCompRowMatrix)LB, bNext, b);
  //}
  try {
    //solverA.solve(LA,a,a);
    //solverB.solve(LB,b,b);
  } 
  catch(Exception e) {
    println("NO CONVERGE");
  }


  /*
  DenseVector temp = a;
   a = aNext;
   aNext = temp;
   temp = b;
   b = bNext;
   bNext = temp;
   */
}

void reaction() {
  minA = 999;
  maxA = -999;
  double valA, valB;
  double currA;
  int index = 0;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      currA = a.get(index);
      valA = deltaT*reactionA(currA,b.get(index),index);
      valB = deltaT*reactionB(currA,b.get(index),index);
      a.add(index,valA);
      b.add(index,valB);
      index++;
      if(currA+valA > maxA) maxA = currA+valA;
      if(currA+valA < minA) minA = currA+valA;
    }
  }
}


double reactionA(double aVal, double bVal,int i) {
  return reactionAGS(aVal,bVal,i);
}

double reactionB(double aVal, double bVal,int i) {
  return reactionBGS(aVal,bVal,i);
}

void setBall(int x, int y, float dist, float aVal, float bVal) {
  int m = int(dist);
  int cx,cy;
  int index;
  for(int i=-m;i<=m;++i) {
    for(int j=-m;j<=m;++j) {
      if(i*i+j*j < dist*dist) {
        cx = (x+i+w)%w; 
        cy = (y+j+h)%h; 
        index = cy*w+cx;
        a.set(index,aVal+random(-.01,.01));
        b.set(index,bVal+random(-.01,.01));
      }
    }
  }
}

void addFrame() {
  loadPixels();
  for(int j=0;j<h;++j) {
    for(int i=0;i<w;++i) {
      volume.setVoxelAt(i,j,time,red(pixels[j*w+i]));
    }
  }
}

void saveMesh() {
  volume.closeSides();
  surface.reset();
  surface.computeSurfaceMesh(mesh,threshold);
  mesh.saveAsSTL(sketchPath("reaction_"+name+".stl"));
}

void setupVolume() {
  volume = new VolumetricHashMap(new Vec3D(w,h,w),w,h,d,.25);
  surface = new HashIsoSurface(volume);
}


void keyPressed() {
  if(key == ' ') {
    //step();
    println(frameCount); 
  } 
  else if(key == 's') {
    saveMesh();
    save("reaction_"+name+".png");
  } 
  else if(key == 'v') {
    view = !view;
  } 
}

