import toxi.geom.*;
import toxi.util.datatypes.*;
import toxi.geom.mesh.*;

import processing.video.*;

import processing.opengl.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.GLU;
import com.sun.opengl.util.*; 


import toxi.volume.*;

import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.io.*;
import no.uib.cipr.matrix.distributed.*;
import no.uib.cipr.matrix.*;

import peasy.*;


DenseVector a, aNext;
DenseVector b, bNext;

Matrix LA;
Matrix LB;
IterativeSolver solverA;
IterativeSolver solverB;

//float[][][] a;
//float[][][] b;
//float[][][] aNext;
//float[][][] bNext;
int w = 64;
int h = 64;
int d = 64;

float delx = .6;
float delxSq = delx*delx;
float deltaT = .5;
float reactionRate = 1;//0.003 //1.0/128.0; //Turing 1/128.0
float aRateX = .04;//.0028 //.1 //.2 //1/16:1/14 OR 1/16:1
float aRateY = .04;
float aRateZ = .05;
float bRate = .01;//.5  //.05//.1 //.25
float F = .06;//.635 //.064//.04 //.06
float k = .06;//.61 // .0606.06 //.062
float EPSILON = .005;

PeasyCam cam;

int iNext[] = new int[w];
int jNext[] = new int[h];
int kNext[] = new int[d];
int iPrev[] = new int[w];
int jPrev[] = new int[h];
int kPrev[] = new int[d];

boolean view = true;
boolean closeMesh = false;

//toxi marching cubes stuff
VolumetricSpace volume;
IsoSurface surface;
TriangleMesh mesh = new WETriangleMesh("why");

float thickness=3.0;

MovieMaker mm;
boolean movie = false;
String name = ""+day()+hour()+minute()+second();
GL gl;
GLU glu;
GLUT glut;
float objScale = 5;
float threshold = .38;

boolean pause=false;
boolean thickened = false;

void setup() {
  size(500,500,OPENGL);
  gl = ((PGraphicsOpenGL) this.g).beginGL();
  colorMode(HSB,1.0);
  a = new DenseVector(w*h*d);
  b = new DenseVector(w*h*d);

  LA = new FlexCompRowMatrix(w*h*d,w*h*d);
  LB = new FlexCompRowMatrix(w*h*d,w*h*d);
  solverA = new CGS(a);
  solverB = new CGS(b);
  cam = new PeasyCam(this, w/2,h/2,d/2,600);

  for(int i=0;i<w;++i) {
    iNext[i] = (i+1)%w;
    iPrev[i] = (i-1+w)%w;
  }
  for(int j=0;j<h;++j) {
    jNext[j] = (j+1)%h;
    jPrev[j] = (j-1+h)%h;
  }
  for(int k=0;k<d;++k) {
    kNext[k] = (k+1)%d;
    kPrev[k] = (k-1+d)%d;
  }
  println("initialize");
  initialize();
  println("load operators");

  aNext = new DenseVector(a);
  bNext = new DenseVector(b);

  if(movie) mm = new MovieMaker(this, width, height, "rxn_"+name+".mov",30,MovieMaker.ANIMATION,MovieMaker.HIGH);
  glSetup();
  setupVolume();
}

void initialize() {
  for(int i=0;i<w*h*d;++i) {
    a.set(i,.5+random(-.01,.01));
    b.set(i,.25+random(-.01,.01));
  }
  for(int l=0;l<10;++l) {
    int x = int(random(w));
    int y = int(random(h));
    int z = int(random(d));
    setBall(x,y,z,8,1.0,0);
  }
  aNext = new DenseVector(a);
  bNext = new DenseVector(b);
  int cw = w/2;
  int ch = h/2;
  int cd = d/2;
  println("compute laplacian");
  computeLaplacian();
}

void draw() {
  glStart();
  noStroke();
  if(!pause) {
    long start = millis();
    while(millis()-start < 30) step();
  }

  if(view) {
    TriangleMesh currMesh;
    getSurface();
    if (thickened) {currMesh = thicken(); currMesh.computeFaceNormals();currMesh.computeVertexNormals();}
    else currMesh = mesh;
    glSetLights();
    glSetMaterials();
    
    int num=currMesh.getNumFaces();
    Vertex v1, v2, v3;
    gl.glBegin(GL.GL_TRIANGLES);
    for(int i=0; i<num; i++) {
      Face f = currMesh.faces.get(i);
      v1 = f.a;
      v2 = f.b;
      v3 = f.c;
      gl.glNormal3f(-v1.normal.x,-v1.normal.y,-v1.normal.z);
      gl.glVertex3f(v1.x,v1.y,v1.z);

      gl.glNormal3f(-v2.normal.x,-v2.normal.y,-v2.normal.z);
      gl.glVertex3f(v2.x,v2.y,v2.z);

      gl.glNormal3f(-v3.normal.x,-v3.normal.y,-v3.normal.z);
      gl.glVertex3f(v3.x,v3.y,v3.z);
    }
    gl.glEnd();
  } 

  else {  
    gl.glColorMaterial(GL.GL_FRONT_AND_BACK, GL.GL_AMBIENT_AND_DIFFUSE);
    gl.glPointSize(8); 
    gl.glBegin(GL.GL_POINTS);

    float valA;
    int index = 0;
    for(int i=0;i<w;++i) {
      for(int j=0;j<h;++j) {
        for(int k=0;k<d;++k) {
          valA = (float)a.get(index++);
          if(valA > threshold) {
            color cc = color((1-valA),1,1);
            gl.glColor4f(red(cc),green(cc),blue(cc),.2);
            gl.glVertex3f((i-w/2)*objScale,(j-h/2)*objScale,(k-d/2)*objScale);
          }
        }
      }
    }
    gl.glEnd();
  }
  if(movie) mm.addFrame();
  ((PGraphicsOpenGL) g).endGL();
}

void step() {
  long start = millis();
  diffusion();
  //println(millis()-start);
  reaction();
}

void diffusion() {
  float valA, valB;
  LA.multAdd(1,a,a);
  LB.multAdd(1,b,b);

  //for(int i=0;i<2;++i) {
  //gaussSeidel((FlexCompRowMatrix)LA, aNext, a);
  //gaussSeidel((FlexCompRowMatrix)LB, bNext, b);
  //}
  try {
    //solverA.solve(LA,a,a);
    //solverB.solve(LB,b,b);
  } 
  catch(Exception e) {
    println("NO CONVERGE");
  }


  DenseVector temp = a;
  a = aNext;
  aNext = temp;
  temp = b;
  b = bNext;
  bNext = temp;
}

void reaction() {
  double valA, valB;
  int index = 0;
  for(int i=0;i<w;++i) {
    for(int j=0;j<h;++j) {
      for(int k=0;k<d;++k) {
        valA = deltaT*reactionA(a.get(index),b.get(index));
        valB = deltaT*reactionB(a.get(index),b.get(index));
        a.add(index,valA);
        b.add(index,valB);
        index++;
      }
    }
  }
}


double reactionA(double aVal, double bVal) {
  return reactionAGS(aVal,bVal);
}

double reactionB(double aVal, double bVal) {
  return reactionBGS(aVal,bVal);
}

void setBall(int x, int y, int z, float dist, float aVal, float bVal) {
  int m = int(dist);
  int cx,cy,cz;
  int index;
  for(int i=-m;i<=m;++i) {
    for(int j=-m;j<=m;++j) {
      for(int k=-m;k<=m;++k) {
        if(i*i+j*j+k*k < dist*dist) {
          cx = (x+i+w)%w; 
          cy = (y+j+h)%h; 
          cz = (z+k+d)%d;
          index = xyzToIndex(cx,cy,cz);          
          a.set(index,aVal+random(-.01,.01));
          b.set(index,bVal+random(-.01,.01));
        }
      }
    }
  }
}

float[] anisotropic(float a1,float a2, float a3, PVector v) {
  float[] terms = new float[6];
  float a1Sq = a1*a1;  
  float a2Sq = a2*a2;  
  float a3Sq = a3*a3;
  float vx1 = v.x+1;
  float vx1Sq = vx1*vx1;
  float vxSq = v.x*v.x; 
  float vySq = v.y*v.y; 
  float vzSq = v.z*v.z;
  terms[0] = a1Sq*vxSq+a2Sq*vySq+a3Sq*vzSq; //1,1
  float c1 = vySq*vzSq/vx1Sq;
  terms[1] = a1Sq*vySq+a2Sq*sq(v.x+vzSq/vx1)+a3Sq*c1; //2,2
  terms[2] = a1Sq*vzSq+a2Sq*c1+a3Sq*sq(v.x+vySq/vx1); //3,3
  terms[3] = (a2Sq-a1Sq)*v.x*v.y + (a2Sq-a3Sq)*v.y*vzSq/vx1; //1,2
  terms[4] = (a3Sq-a1Sq)*v.x*v.z + (a3Sq-a2Sq)*vySq*v.z/vx1; //1,3
  terms[5] = a1Sq*v.y*v.z-(a2Sq+a3Sq)*v.x*v.y*v.z/vx1-(a2Sq*vzSq+a3Sq*vySq)*v.y*v.z/vx1Sq;
  return terms;
}

TriangleMesh thicken() {
  TriangleMesh thickMesh = new TriangleMesh("thisisdumb");
  WEFace f;
  Vec3D aUp, aDown, bUp, bDown, cUp,cDown;
  for(int i=0;i<mesh.faces.size();++i) {
    f = (WEFace) mesh.faces.get(i);
    aUp = f.a.add(f.a.normal.scale(thickness/2));
    aDown = f.a.sub(f.a.normal.scale(thickness/2));
    bUp = f.b.add(f.b.normal.scale(thickness/2));
    bDown = f.b.sub(f.b.normal.scale(thickness/2));
    cUp = f.c.add(f.c.normal.scale(thickness/2));
    cDown = f.c.sub(f.c.normal.scale(thickness/2));
    thickMesh.addFace(aUp,bUp,cUp);
    thickMesh.addFace(cDown,bDown,aDown);

    //check if edge is on the boundary (only one neighboring face)
    if(f.edges.get(0).faces.size() == 1) {
      //if so close that edge
      thickMesh.addFace(bUp,aUp,aDown);
      thickMesh.addFace(aDown,bDown,bUp);
    }
    if(f.edges.get(1).faces.size() == 1) {
      thickMesh.addFace(cUp,bUp,bDown);
      thickMesh.addFace(bDown,cDown,cUp);
    }
    if(f.edges.get(2).faces.size() == 1) {
      thickMesh.addFace(aUp,cUp,cDown);
      thickMesh.addFace(cDown,aDown,aUp);
    }
  }
  return thickMesh;
}

void setupVolume() {
  volume = new VolumetricSpaceArray(new Vec3D(w,h,d),w,h,d);
  surface = new ArrayIsoSurface(volume);
}

void getSurface() {
  float[] volumeData = ((VolumetricSpaceArray)volume).getData();
  int index = 0;

  for(int z=0; z<d; z++) {
    for(int y=0; y<h; y++) {  
      for(int x=0; x<w; x++) {
        volumeData[index] = (float)a.get(xyzToIndex(x,y,z));
        index++;
      }
    }
  }
  if (closeMesh) volume.closeSides();
  surface.reset();
  surface.computeSurfaceMesh(mesh,threshold);
  mesh.computeVertexNormals();
  mesh.scale(objScale);
}

void keyPressed() {
  if(key == 't') thickened = !thickened;
  if(key == 'v') view = !view;
  else if(key == 's') {
    getSurface();
    
    thicken().saveAsSTL(sketchPath("reaction-"+day()+"_"+hour()+"_"+minute()+".stl"));
  } 
  else if(key == 'f') {
    if(movie) mm.finish();
    movie = false;
  } 
  else if(key == ' ') step();
  else if(key=='p') pause=!pause;
  else if(key=='c') closeMesh = !closeMesh;
}

