import toxi.geom.*;
import toxi.geom.mesh.*;

import no.uib.cipr.matrix.sparse.*;
import no.uib.cipr.matrix.*;

import peasy.org.apache.commons.math.*;
import peasy.*;
import peasy.org.apache.commons.math.geometry.*;

import processing.opengl.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.GLU;

String meshFile = "sphereSmall3.stl";

TriangleMesh mesh, offsetMesh;
STLReader reader = new STLReader();
GL gl;
GLU glu;

PeasyCam cam;

float reactionRate = 2;
float diffA = 2;
float diffB = .5;
float F = .06; //.04
float k = .06; //.06
float delT = .1;
float objScale = 300;
float offset = 150;

int numVertices;

FlexCompRowMatrix laplacian;
DenseVector aVals;
DenseVector bVals;
DenseVector aVals2;
DenseVector bVals2;

boolean paused = false;
boolean view = false;
int rdIt = 30;
String name = ""+day()+hour()+minute()+second();

void setup() {
  size(500,500,OPENGL);
  gl = ((PGraphicsOpenGL) this.g).beginGL();
  cam = new PeasyCam(this,0,0,0,900);
  loadMesh();
  laplacian = new FlexCompRowMatrix(numVertices, numVertices);
  aVals = new DenseVector(numVertices);
  bVals = new DenseVector(numVertices);
  computeLaplacian();
  glSetup();
  
  for(int i=0;i<numVertices;++i) {
    aVals.set(i,.5+random(-.01,.01));
    bVals.set(i,.25+random(-.01,.01));
  }
  aVals2 = new DenseVector(aVals);
  bVals2 = new DenseVector(bVals);
}

void draw() {
  cam.rotateX(0.0);
  glStart();
  if(!paused) step();
  if(view) drawOffset();
  else drawMesh();    
  ((PGraphicsOpenGL) g).endGL(); 
}

void drawOffset() {
  glSetMaterials();
  glSetLights();
  gl.glBegin(GL.GL_TRIANGLES);
  Vertex v1, v2, v3;
  int i1,i2,i3;
  //Face face;
  gl.glColor3f(.7,.7,.7);
  //for(int i =0;i<m.faces.size();++i) {
  for(Face face : offsetMesh.faces) {
    //face =  m.faces.get(i);
    v1 = face.a;
    v2 = face.b;
    v3 = face.c;
    gl.glNormal3f(v1.normal.x,v1.normal.y,v1.normal.z);
    gl.glVertex3f(v1.x,v1.y,v1.z);
    
    gl.glNormal3f(v2.normal.x,v2.normal.y,v2.normal.z);
    gl.glVertex3f(v2.x,v2.y,v2.z);
    
    gl.glNormal3f(v3.normal.x,v3.normal.y,v3.normal.z);
    gl.glVertex3f(v3.x,v3.y,v3.z);
  }
  gl.glEnd();

}

void drawMesh() {
  gl.glBegin(GL.GL_TRIANGLES);
  Vertex v1, v2, v3;
  int i1,i2,i3;
  float aVal;
  //Face face;
  gl.glColor3f(.7,.7,.7);
  //for(int i =0;i<m.faces.size();++i) {
  for(Face face : mesh.faces) {
    //face =  m.faces.get(i);
    v1 = face.a;
    v2 = face.b;
    v3 = face.c;
    aVal = (float) aVals.get(v1.id);
    //gl.glColor3f(v1.a,v1.a,v1.a);
    gl.glColor3f(aVal,aVal,aVal);
    gl.glNormal3f(v1.normal.x,v1.normal.y,v1.normal.z);
    gl.glVertex3f(v1.x,v1.y,v1.z);
    
    aVal = (float) aVals.get(v2.id);    
    gl.glColor3f(aVal,aVal,aVal);
    //gl.glColor3f(v1.a,v1.a,v1.a);
    gl.glNormal3f(v2.normal.x,v2.normal.y,v2.normal.z);
    gl.glVertex3f(v2.x,v2.y,v2.z);
    
    aVal = (float) aVals.get(v3.id);
    gl.glColor3f(aVal,aVal,aVal);
    //gl.glColor3f(v1.a,v1.a,v1.a);
    gl.glNormal3f(v3.normal.x,v3.normal.y,v3.normal.z);
    gl.glVertex3f(v3.x,v3.y,v3.z);
  }
  gl.glEnd();

}

void loadMesh() {
  mesh = (TriangleMesh) reader.loadBinary(dataPath(meshFile), STLReader.TRIANGLEMESH);
  mesh.scale(objScale);
  numVertices = mesh.getNumVertices();
  mesh.computeFaceNormals();
  mesh.computeVertexNormals();
}

void step() {
  
  for(int i=0;i<rdIt;++i) {
    diffusion();
    if(checkNaN()) {
      println(i);
      paused = true;
      return;
    }
    react();
    
    if(paused) {
      println(i);
      return;
    }
  }
}

void react() {
  float na, nb;
  float aVal,bVal;
  for(int i=0;i<numVertices;++i) {
    aVal = (float) aVals.get(i);
    bVal = (float) bVals.get(i);
    if(Float.isNaN(aVal) || Float.isNaN(bVal)) {
      println("REACT NAN " + i);
      paused = true;
    }
    na = delT*reactionGSA(aVal,bVal);  
    nb = delT*reactionGSB(aVal,bVal);
    //vertices[i].a += na;
    //vertices[i].b += nb;
    aVals.add(i,na);
    bVals.add(i,nb);
  }
}


//can we do this implicitly??
void diffusion() {
  laplacian.multAdd(delT*diffA, aVals, aVals);
  laplacian.multAdd(delT*diffB, bVals, bVals);  
}

void computeLaplacian() {
  Vertex v1,v2,v3;
  int i1,i2,i3;
  float area;
  Vec3D cross,d1,d2,d3;
  float sinMag, cos1,cos2,cos3, sin1,sin2,sin3;
  float dist12;
  float dist13;
  float dist23;
  laplacian.zero();
  //for(int i=0;i<faces.size();++i) {
  for(Face face : mesh.faces) {
    //face = (int[]) faces.get(i);
    v1 = face.a;
    v2 = face.b;
    v3 = face.c;
    i1 = v1.id;
    i2 = v2.id;
    i3 = v3.id;
    d1 = v2.sub(v1);
    d2 = v3.sub(v1);
    d3 = v3.sub(v2);
    dist12 = d1.magnitude();
    dist13 = d2.magnitude();
    dist23 = d3.magnitude();
    
    
    cross = d1.cross(d2);
    sinMag = cross.magnitude();

    sin1 = sinMag/(dist12*dist13);
    sin2 = sinMag/(dist12*dist23);
    sin3 = sinMag/(dist13*dist23);
    
    cos1 = d1.x*d2.x+d1.y*d2.y+d1.z*d2.z;
    cos2 = -d1.x*d3.x-d1.y*d3.y-d1.z*d3.z;
    cos3 = d3.x*d2.x+d3.y*d2.y+d3.z*d2.z;
    cos1 /= dist12*dist13;
    cos2 /= dist12*dist23;
    cos3 /= dist13*dist23;
    
    laplacian.add(i1,i2,sin1/((cos1+1)*dist12));
    laplacian.add(i2,i1,sin2/((cos2+1)*dist12));
    laplacian.add(i1,i3,sin1/((cos1+1)*dist13));
    laplacian.add(i3,i1,sin3/((cos3+1)*dist13));
    laplacian.add(i2,i3,sin2/((cos2+1)*dist23));
    laplacian.add(i3,i2,sin3/((cos3+1)*dist23));
    
    laplacian.add(i1,i1,-sin1/((cos1+1)*dist12));
    laplacian.add(i2,i2,-sin2/((cos2+1)*dist12));
    laplacian.add(i1,i1,-sin1/((cos1+1)*dist13));
    laplacian.add(i3,i3,-sin3/((cos3+1)*dist13));
    laplacian.add(i2,i2,-sin2/((cos2+1)*dist23));
    laplacian.add(i3,i3,-sin3/((cos3+1)*dist23));

  }
}

float reactionGSA(float aVal, float bVal) {
  return reactionRate*(-aVal*bVal*bVal+F*(1-aVal));
}

float reactionGSB(float aVal, float bVal) {
  return reactionRate*(aVal*bVal*bVal-(F+k)*bVal);
}

class rdVertex extends PVector {
  float a,b;
  float newA;
  float newB;
  float area;
  PVector norm;
  
  rdVertex(float x,float y, float z) {
    super(x,y,z);
    a = 0;
    b = 0;
  }
}

void keyPressed() {
  if(key == ' ') {
    step();
  } else if(key == 's') {
    saveFrame("meshRD-####.png");
  } else if(key == 'p') {
    paused = !paused;
  } else if(key == 'e') {
    offsetExport();
  } else if(key == 'v') {
    view = !view;
    offsetMesh();
  }
}

float distSq(PVector v1, PVector v2) {
  return sq(v1.x-v2.x)+sq(v1.y-v2.y)+sq(v1.z-v2.z);
}

void offsetExport() {
  offsetMesh();
  offsetMesh.saveAsSTL(sketchPath("offset_"+name+".stl"));
}

void offsetMesh() {
  offsetMesh = new TriangleMesh();
  Vertex v1,v2,v3;
  Vec3D o1,o2,o3;
  float a1,a2,a3;
  for(Face face : mesh.faces) {
    v1 = face.a;
    v2 = face.b;
    v3 = face.c;
    a1 = (float) aVals.get(v1.id);
    a2 = (float) aVals.get(v2.id);
    a3 = (float) aVals.get(v3.id);
    //offset out
    o1 = v1.add(v1.normal.scale(a1*offset));
    o2 = v2.add(v2.normal.scale(a2*offset));
    o3 = v3.add(v3.normal.scale(a3*offset));
    offsetMesh.addFace(o1,o2,o3);

    //offset in
    o1 = v1.sub(v1.normal.scale(a1*offset));
    o2 = v2.sub(v2.normal.scale(a2*offset));
    o3 = v3.sub(v3.normal.scale(a3*offset));
    offsetMesh.addFace(o1,o2,o3);
  }  
}

boolean checkNaN() {
  float bVal, aVal;
  for(int i=0;i<numVertices;++i) {
    aVal = (float) aVals.get(i);
    bVal = (float) bVals.get(i);
    if(Float.isNaN(aVal) || Float.isNaN(bVal)) {
      //println(laplacian);
      return true;
    }
  }
  return false;
}
